use std::{str::FromStr, sync::Arc};

use derive_more::derive::Into;
use thiserror::Error;

use crate::constants;
#[derive(Debug, Error, PartialEq, Eq)]
pub enum InvalidGivenTaskError {
    #[error(
        "({0}) is not a valid format for unsigned whole number between 1 and {}",
        constants::MAX_TASK
    )]
    InvalidFormat(Arc<str>),
    #[error(
        "({0}) is within the valid range between 1 and {}",
        constants::MAX_TASK
    )]
    InvalidRange(u32),
}

#[derive(Debug, Copy, Into, Clone, PartialEq, Eq)]
pub struct GivenTask(u32);

impl GivenTask {
    pub fn new(value: u32) -> Result<Self, InvalidGivenTaskError> {
        if value < 1 || value > constants::MAX_TASK {
            Err(InvalidGivenTaskError::InvalidRange(value))
        } else {
            Ok(Self(value))
        }
    }
}

impl FromStr for GivenTask {
    type Err = InvalidGivenTaskError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let number: u32 = s
            .parse()
            .map_err(|_| InvalidGivenTaskError::InvalidFormat(Arc::from(s)))?;

        GivenTask::new(number)
    }
}

#[cfg(test)]
mod testing {
    use crate::{
        cli::GivenDay,
        constants::{self, MAX_DAY},
    };

    use super::*;

    #[test]
    fn parse_correct_value() {
        fn assert_case(input: &str, expected: u32) {
            let actual = GivenDay::from_str(input).unwrap();
            assert_eq!(expected, actual.into(), "input: {}", input);
        }

        assert_case("1", 1);
        assert_case(&constants::MAX_DAY.to_string(), MAX_DAY);
    }

    #[test]
    fn detect_invalid_input() {
        fn assert_case(input: &str, expected_err: InvalidGivenTaskError) {
            let actual = GivenTask::from_str(input).unwrap_err();
            assert_eq!(expected_err, actual, "input: {}", input);
        }

        const OVER_MAX: u32 = constants::MAX_TASK + 1;
        assert_case("0", InvalidGivenTaskError::InvalidRange(0));
        assert_case(
            &OVER_MAX.to_string(),
            InvalidGivenTaskError::InvalidRange(OVER_MAX),
        );
        assert_case("", InvalidGivenTaskError::InvalidFormat("".into()));
        assert_case("a", InvalidGivenTaskError::InvalidFormat("a".into()));
    }
}
