mod joker_ordered;
pub use joker_ordered::JokerOrdered;

use crate::day7::second_ordering;
use std::usize;

use super::{dealt_hand::DealtHand, hand_kind::HandKind, second_ordering::StrengthOfSymbols};

#[derive(Debug, PartialEq, Eq)]
pub struct CategorizedHand {
    kind: HandKind,
    rest: DealtHand,
}

impl CategorizedHand {
    #[cfg(test)]
    pub fn new(kind: HandKind, rest: DealtHand) -> Self {
        Self { kind, rest }
    }

    pub fn rest(&self) -> &DealtHand {
        &self.rest
    }
    pub fn bid(&self) -> usize {
        self.rest().bid()
    }
}

impl Ord for CategorizedHand {
    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
        hand_cmp(self, other, &second_ordering::LABEL_ORDERING)
    }
}

fn hand_cmp(
    left: &CategorizedHand,
    right: &CategorizedHand,
    label_ordering: &StrengthOfSymbols,
) -> std::cmp::Ordering {
    match left.kind.cmp(&right.kind) {
        std::cmp::Ordering::Equal => {
            assert_eq!(
                left.kind, right.kind,
                "If ordering is equal then the hand kind must be the same too\n\
                    left kind: ({:?}) and right kind: ({:?})",
                left.kind, right.kind
            );
            second_ordering::compare_along_str(
                left.rest().hand(),
                right.rest().hand(),
                label_ordering,
            )
        }
        not_equal_ordering => {
            assert_ne!(
                left.kind, right.kind,
                "If ordering is not equal then the hand kind must not be the same either\n\
                    left kind: ({:?}) and right kind: ({:?})",
                left.kind, right.kind
            );
            not_equal_ordering
        }
    }
}

impl PartialOrd for CategorizedHand {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        Some(self.cmp(other))
    }
}

impl From<DealtHand> for CategorizedHand {
    fn from(value: DealtHand) -> Self {
        let kind = value.hand().into();
        CategorizedHand { kind, rest: value }
    }
}
impl From<(DealtHand, char)> for CategorizedHand {
    fn from((value, joker): (DealtHand, char)) -> Self {
        let kind = (value.hand(), joker).into();
        CategorizedHand { kind, rest: value }
    }
}

#[cfg(test)]
mod testing {
    use std::cmp::Ordering;

    use crate::day7::{categorized_hand::CategorizedHand, dealt_hand::DealtHand};

    #[test]
    fn compare_correctly_by_kind_and_second_ordering() {
        fn assert_case(left: DealtHand, right: DealtHand, expected: Ordering) {
            let (left, right): (CategorizedHand, CategorizedHand) = (left.into(), right.into());
            let actual = left.cmp(&right);
            assert_eq!(expected, actual, "Left: {:?}\nRight: {:?}", left, right);
        }

        assert_case(
            DealtHand::new("AAAAA", 20),
            DealtHand::new("BBBBA", 30),
            Ordering::Greater,
        );
        assert_case(
            DealtHand::new("12345", 20),
            DealtHand::new("BBBBA", 30),
            Ordering::Less,
        );
        assert_case(
            DealtHand::new("12345", 20),
            DealtHand::new("28495", 30),
            Ordering::Less,
        );
        assert_case(
            DealtHand::new("QQQJA", 20),
            DealtHand::new("T55J5", 30),
            Ordering::Greater,
        );
        //KK677 and KTJJT are both two pair.
        //Their first cards both have the same label,
        //but the second card of KK677 is stronger (K vs T), so KTJJT gets rank 2 and KK677 gets rank 3.
        assert_case(
            DealtHand::new("KK677", 20),
            DealtHand::new("KTJJT", 30),
            Ordering::Greater,
        );

        //So, 33332 and 2AAAA are both four of a kind hands,
        //but 33332 is stronger because its first card is stronger.
        assert_case(
            DealtHand::new("33332", 20),
            DealtHand::new("2AAAA", 30),
            Ordering::Greater,
        );

        //Similarly, 77888 and 77788 are both a full house,
        //but 77888 is stronger because its third card is stronger
        //(and both hands have the same first and second card).
        assert_case(
            DealtHand::new("77788", 30),
            DealtHand::new("77888", 20),
            Ordering::Less,
        );
        assert_case(
            DealtHand::new("19231", 30),
            DealtHand::new("18231", 20),
            Ordering::Greater,
        );
    }

    #[test]
    fn sort_example_correctly() {
        //32T3K is the only one pair and the other hands are all a stronger type, so it gets rank 1.
        //
        //KK677 and KTJJT are both two pair.
        //Their first cards both have the same label,
        //but the second card of KK677 is stronger (K vs T), so KTJJT gets rank 2 and KK677 gets rank 3.
        //
        //T55J5 and QQQJA are both three of a kind.
        //QQQJA has a stronger first card, so it gets rank 5 and T55J5 gets rank 4.

        let mut input: Vec<_> = vec![
            DealtHand::new("32T3K", 765),
            DealtHand::new("T55J5", 684),
            DealtHand::new("KK677", 28),
            DealtHand::new("KTJJT", 220),
            DealtHand::new("QQQJA", 483),
        ]
        .into_iter()
        .map(CategorizedHand::from)
        .collect();

        input.sort();

        insta::assert_debug_snapshot!(input);
    }
}
